import math
import re


def Add(number1, number2):
    """
    Adds two numbers together and rounds the result if it is not a whole number.
    @param number1: The first number.
    @param number2: The second number.
    @return: The sum of the two numbers, rounded to the nearest decimal if necessary.
    """
    result = number1 + number2
    # rounds result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def Subtract(number1, number2):
    """
    Subtracts the second number from the first and rounds the result if it is not a whole number.
    @param number1: The first number.
    @param number2: The number to subtract from the first number.
    @return: The difference of the two numbers, rounded to the nearest decimal if necessary.
    """
    result = number1 - number2
    # round result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def Multiply(number1, number2):
    """
    Multiplies two numbers together and rounds the result if it is not a whole number.
    @param number1: The first number.
    @param number2: The second number.
    @return: The product of the two numbers, rounded to the nearest decimal if necessary.
    """
    result = number1 * number2
    # round result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def Divide(number1, number2):
    """
    Divides the first number by the second and rounds the result if it is not a whole number.
    Raises an error if the divisor is zero.
    @param number1: The dividend.
    @param number2: The divisor.
    @return: The quotient of the two numbers, rounded to the nearest decimal if necessary.
    @raises ZeroDivisionError: If the second number (divisor) is zero.
    """
    if number2 == 0:
        raise ZeroDivisionError("Error: Division by Zero")
    result = number1 / number2
    # round result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def Factorial(number):
    """
    Calculates the factorial of a non-negative integer.
    @param number: The number to calculate the factorial of.
    @return: The factorial of the number.
    @raises ValueError: If the number is not a non-negative integer.
    """
    if (not isinstance(number, int) and int(number) != number) or number < 0:
        raise ValueError("Factorial is defined for non-negative integers")
    # using math.factorial function to compute factorial
    return math.factorial(int(number))


def Exponentiate(base, exponent):
    """
    Raises a base to the power of an exponent and rounds the result if it is not a whole number.
    @param base: The base number.
    @param exponent: The exponent.
    @return: The result of the exponentiation, rounded to the nearest decimal if necessary.
    """
    result = base ** exponent
    # round result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def Root(number, root=2):
    """
    Calculates the nth root of a number, checking if the operation is defined (e.g., no even roots of negative numbers).
    Defaulting the root to 2.
    @param number: The number to find the root of.
    @param root: The root to calculate.
    @return: The nth root of the number, rounded to the nearest decimal if necessary.
    @raises ValueError: If an even root is attempted on a negative number.
    """
    if number < 0 and root % 2 == 0:
        raise ValueError("Even roots are not defined for negative numbers")
    result = number ** (1 / root)
    # round result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def Sine(angle_degrees):
    """
    Calculates the sine of an angle in degrees.
    @param angle_degrees: The angle in degrees.
    @return: The sine of the angle, rounded to the nearest decimal if necessary.
    """
    angle_radians = math.radians(angle_degrees)
    result = math.sin(angle_radians)
    # round result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def Cosine(angle_degrees):
    """
    Calculates the cosine of an angle in degrees.
    @param angle_degrees: The angle in degrees.
    @return: The cosine of the angle, rounded to the nearest decimal if necessary.
    """
    angle_radians = math.radians(angle_degrees)
    result = math.cos(angle_radians)
    # round result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def Logarithm(value, base=10):
    """
    Calculates the logarithm of a value to a specified base, defaulting to base 10.
    @param value: The value to calculate the logarithm of.
    @param base: The base of the logarithm.
    @return: The logarithm of the value to the specified base, rounded to the nearest decimal if necessary.
    @raises ValueError: If the value is non-positive.
    """
    if value <= 0:
        raise ValueError("Logarithm is not defined for non-positive values.")
    result = math.log(value, base)
    # round result if it is a float
    return round(result, 10) if result % 1 != 0 else int(result)


def ConvertToRPN(expression):
    """
    Converts a mathematical expression in infix notation to Reverse Polish Notation (RPN) using the shunting yard
    algorithm.
    @param expression: The infix expression to convert.
    @return: The converted expression in RPN.
    """
    # define precedence and associativity of operators
    precedence = {
        '+': 2,
        '-': 2,
        '*': 1,
        '/': 1,
        '^': 0,
        'sin': -1,  # functions have the lowest precedence
        'cos': -1,
        'log': -1,
        'sqrt': -1,
        '!': 0
    }
    right_associative = {'^', '!', 'sin', 'cos', 'tan', 'log', 'sqrt'}

    # tokenize the input expression
    def tokenize(expr):
        # extract tokens using regex
        tokens = re.findall(r'\b[a-zA-Z]+\b|[\d\.]+|[-+/*^()]|!', expr.replace(' ', ''))
        new_tokens = []
        i = 0
        while i < len(tokens):
            token = tokens[i]
            if token == '-' and (i == 0 or tokens[i - 1] in '(-+*/^'):
                if i + 1 < len(tokens) and tokens[i + 1].replace('.', '', 1).isdigit():
                    new_tokens.append('-' + tokens[i + 1])  # combine unary minus with the number
                    i += 1  # skip the next token as it now is part of the current token
                else:
                    new_tokens.append('u-')  # unary minus for negating expressions (will be removed later)
            else:
                new_tokens.append(token)
            i += 1
        return new_tokens

    # convert tokens from infix to postfix notation (RPN)
    def shunting_yard(tokens):
        output = []
        operators = []
        for token in tokens:
            if token.isnumeric() or re.match(r'-?\d+(\.\d+)?', token):  # if it is a number (including negative)
                output.append(token)
            elif token in precedence:
                while (operators and operators[-1] != '(' and
                       (precedence[operators[-1]] < precedence[token] or
                        (precedence[operators[-1]] == precedence[token] and token not in right_associative))):
                    output.append(operators.pop())
                operators.append(token)
            elif token == '(':
                operators.append(token)
            elif token == ')':
                while operators and operators[-1] != '(':
                    output.append(operators.pop())
                operators.pop()

        while operators:
            output.append(operators.pop())

        return output

    tokens = tokenize(expression)
    rpn = shunting_yard(tokens)
    return ' '.join(rpn)


def EvaluateString(expression):
    """
    Evaluates a mathematical expression in infix notation and returns the result.
    @param expression: The expression to evaluate.
    @return: The result of the evaluated expression.
    """
    # check for invalid function usage
    if re.search(r'\b(sin|cos|sqrt|log)\s+(sin|cos|sqrt|log)', expression):
        raise ValueError("Invalid function usage")

    # handle empty input
    if expression == "":
        return 0

    # convert expression to RPN and evaluate
    rpn_expression = ConvertToRPN(expression)
    tokens = rpn_expression.split()
    stack = []

    for token in tokens:
        if token.isdigit() or re.match(r'-?\d+(\.\d+)?', token):  # checking if the token is a number
            stack.append(float(token))
        elif token in {'+', '-', '*', '/', '^', 'sqrt', 'sin', 'cos', 'log', '!'}:
            if len(stack) < 2 and token not in {'sin', 'cos', 'log', 'sqrt', '!'}:
                raise ValueError("Insufficient values in the expression.")
            if token == '+':
                b = stack.pop()
                a = stack.pop()
                stack.append(Add(a, b))
            elif token == '-':
                b = stack.pop()
                a = stack.pop()
                stack.append(Subtract(a, b))
            elif token == '*':
                b = stack.pop()
                a = stack.pop()
                stack.append(Multiply(a, b))
            elif token == '/':
                b = stack.pop()
                a = stack.pop()
                stack.append(Divide(a, b))
            elif token == '^':
                b = stack.pop()
                a = stack.pop()
                stack.append(Exponentiate(a, b))
            elif token == 'sqrt':
                a = stack.pop()
                stack.append(Root(a))  # default root is 2 and in the gui we only have square root implemented
            elif token == 'sin':
                a = stack.pop()
                stack.append(Sine(a))
            elif token == 'cos':
                a = stack.pop()
                stack.append(Cosine(a))
            elif token == 'log':
                a = stack.pop()
                stack.append(Logarithm(a))  # default base is 10 and in the gui we only have log base 10 implemented
            elif token == '!':
                a = stack.pop()
                stack.append(Factorial(a))
        else:
            raise ValueError(f"Unexpected token: {token}")

    return int(stack[0] if stack else 0) \
        if int(stack[0] if stack else 0) == (stack[0] if stack else 0) \
        else (stack[0] if stack else 0)
