import sys

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFontDatabase, QKeyEvent
from PyQt6.QtWidgets import QApplication, QWidget, QVBoxLayout, QPushButton, QLineEdit, QGridLayout

from calc import EvaluateString


class CalculatorGUI(QWidget):
    def __init__(self):
        super().__init__()
        self.current_input = ""
        self.operation_buffer = []
        self.error_state = False
        self.setMinimumSize(500, 500)
        QFontDatabase.addApplicationFont("src/calculator/InriaSans-Bold.ttf")
        self.initUI()

    def initUI(self):
        self.setWindowTitle('Calculator')
        mainLayout = QVBoxLayout()
        self.setLayout(mainLayout)
        self.layout().setSpacing(0)
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.display = QLineEdit()
        self.display.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.display.setReadOnly(True)
        self.display.setStyleSheet(
            "QLineEdit { background-color: #404040; color: white; font-size: 32px; border: none; padding: 20px; }")
        mainLayout.addWidget(self.display)

        button_layout = QGridLayout()
        button_layout.setSpacing(0)
        button_layout.setContentsMargins(0, 0, 0, 0)
        self.setupButtons(mainLayout, button_layout)

    def setupButtons(self, mainLayout, button_layout):
        button_specs = [
            ('sin', 'cos', 'xˆ2', '√x', 'AC', '#3E3E3E', ['#3E3E3E', '#3E3E3E', '#3E3E3E', '#3E3E3E', '#FF5858']),
            ('7', '8', '9', '*', '/', '#595959', ['#595959', '#595959', '#595959', '#35BCBC', '#35BCBC']),
            ('4', '5', '6', '-', '+/-', '#595959', ['#595959', '#595959', '#595959', '#35BCBC', '#35BCBC']),
            ('1', '2', '3', '+', '!', '#595959', ['#595959', '#595959', '#595959', '#35BCBC', '#35BCBC']),
            ('0', '.', '=', '%', 'log', '#595959', ['#595959', '#595959', '#199393', '#35BCBC', '#35BCBC'])
        ]

        for row_index, row in enumerate(button_specs):
            labels = row[:-2]
            row_colors = row[-1]
            bg_color = row[-2]
            for col_index, btnText in enumerate(labels):
                if btnText:
                    button = QPushButton(btnText)
                    button.setObjectName(btnText)
                    button.clicked.connect(self.onButtonClick)
                    color = row_colors[col_index] if row_colors[col_index] else bg_color
                    button.setStyleSheet(
                        f"QPushButton {{ background-color: {color}; color: white; font-size: 24px; border: 1px solid "
                        f"black; }}")
                    button_layout.addWidget(button, row_index, col_index)
        mainLayout.addLayout(button_layout)

    def clearInput(self, button_text):
        self.current_input = ""
        self.error_state = False
        if button_text == 'AC':
            self.operation_buffer = []
        self.updateDisplay('0')
        self.updateClearButton()

    def onButtonClick(self):
        button = self.sender()
        buttonText = button.text()
        if buttonText == 'AC' or buttonText == 'C':
            self.clearInput(buttonText)
        else:
            if self.error_state:
                self.clearInput('AC')
            self.processInput(buttonText)

    def processInput(self, input):
        if input.isdigit() or input == '.':
            if input == '.' and '.' in self.current_input:
                return
            self.current_input += input
        elif input in {'+', '-', '*', '/'}:
            if self.current_input:
                self.operation_buffer.append(self.current_input)
                self.operation_buffer.append(input)
                self.current_input = ""
        elif input == '=':
            if self.operation_buffer:
                self.operation_buffer.append(self.current_input)
                expression = ''.join(self.operation_buffer)
                try:
                    self.current_input = str(EvaluateString(expression))
                except Exception as e:
                    self.current_input = "Error"
                    self.error_state = True
                self.operation_buffer = []
        elif input in {'sin', 'cos', 'log', 'xˆ2', '√x', '!'}:
            try:
                expression = self.current_input
                if input in {'sin', 'cos', 'log'}:
                    expression = input + "(" + self.current_input + ")"
                elif input == 'xˆ2':
                    expression = self.current_input + "^2"
                elif input == '√x':
                    expression = 'sqrt(' + self.current_input + ')'
                elif input == '!':
                    expression = self.current_input + "!"
                self.current_input = str(EvaluateString(expression))
            except Exception as e:
                self.current_input = "Error"
                self.error_state = True
        elif input == '+/-':
            if self.current_input:
                self.current_input = '-' + self.current_input if not self.current_input.startswith(
                    '-') else self.current_input[1:]
        elif input == '%':
            try:
                self.current_input = str(float(self.current_input) / 100)
            except ValueError:
                self.current_input = "Error"
                self.error_state = True
        self.updateDisplay(self.current_input)
        self.updateClearButton()

    def updateDisplay(self, text):
        self.display.setText(text if text != "" else "0")

    def updateClearButton(self):
        clear_button = self.findChild(QPushButton, 'AC')
        if self.current_input or self.operation_buffer or self.error_state:
            clear_button.setText('C')
        else:
            clear_button.setText('AC')

    def keyPressEvent(self, event: QKeyEvent):
        key = event.key()
        key_map = {
            Qt.Key.Key_0: '0',
            Qt.Key.Key_1: '1',
            Qt.Key.Key_2: '2',
            Qt.Key.Key_3: '3',
            Qt.Key.Key_4: '4',
            Qt.Key.Key_5: '5',
            Qt.Key.Key_6: '6',
            Qt.Key.Key_7: '7',
            Qt.Key.Key_8: '8',
            Qt.Key.Key_9: '9',
            Qt.Key.Key_Plus: '+',
            Qt.Key.Key_Minus: '-',
            Qt.Key.Key_Asterisk: '*',
            Qt.Key.Key_Slash: '/',
            Qt.Key.Key_Percent: '%',
            Qt.Key.Key_Exclam: '!',
            Qt.Key.Key_Period: '.',
            Qt.Key.Key_Enter: '=',
            Qt.Key.Key_Return: '=',
            Qt.Key.Key_Backspace: 'AC',
            Qt.Key.Key_Equal: '=',
            Qt.Key.Key_Escape: 'AC'
        }
        if key in key_map:
            button_name = key_map[key]
            button = self.findChild(QPushButton, button_name)
            if button:
                button.click()

    def resizeEvent(self, event):
        width = self.width()
        height = int(width * 1.2)  # maintain a 5:6 aspect ratio
        self.setFixedSize(width, height)

        new_size = min(width // 5, height // 6)
        font_size = max(10, new_size // 3)
        self.display.setFixedHeight(new_size)
        self.display.setStyleSheet(
            f"QLineEdit {{ background-color: #404040; color: white; font-size: {font_size}px; border: none; padding: "
            f"20px; }}")
        for button in self.findChildren(QPushButton):
            button.setFixedSize(new_size, new_size)
            button.setStyleSheet(
                f"QPushButton {{ background-color: {button.styleSheet().split(';')[0].split(' ')[-1]}; color: white; "
                f"font-size: {font_size}px; border: 1px solid black; }}")


def main():
    app = QApplication(sys.argv)
    ex = CalculatorGUI()
    ex.show()
    sys.exit(app.exec())


if __name__ == '__main__':
    main()
